#include <glibmm/i18n.h>
#include <gtkmm.h>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <sstream>

#include <boost/format.hpp>

#include "file_utils.hpp"
#include "util.hpp"
#include "paths.hpp"
#include "main.hpp"

#include "amazon.hpp"
#include "musicbrainz_xml.hpp"
#include "library_ui_modify_tracks.hpp"

#include "x_library.hpp"
#include "x_system_control.hpp"
#include "x_vfs.hpp"

namespace Bmp
{
      LibraryUIModifyTracks::LibraryUIModifyTracks (BaseObjectType *cobject,
		                                                const Glib::RefPtr<Gnome::Glade::Xml> &xml)
          : Gtk::Dialog (cobject),
        	  ref_xml (xml),
        	  disc (Gdk::Pixbuf::create_from_file (Glib::build_filename (DATA_DIR, "images" G_DIR_SEPARATOR_S "audio.png")))
      {
        ref_xml->get_widget ("view_release",	view_release);
        ref_xml->get_widget ("view_local",	view_local);
        ref_xml->get_widget ("b_move_up",	b_move_up);
        ref_xml->get_widget ("b_move_down",	b_move_down);
        ref_xml->get_widget ("i_coverart",	i_coverart);

        ref_xml->get_widget ("l_artist", l_artist);
        ref_xml->get_widget ("l_album", l_album);

        b_move_up->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyTracks::on_button_up));
        b_move_down->signal_clicked().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyTracks::on_button_down));

        struct { 
            char    *title;
            double   xalign;
            int	     width;
        } headers_tracks[] = {
            { _("#"),	  1.0, 0,  },
            { _("Title"), 0.0, 100 },
        };

        // Release
        for (unsigned int n = 0; n < G_N_ELEMENTS (headers_tracks); ++n)
        {
          Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
          cell->property_xalign() = headers_tracks[n].xalign;
          view_release->append_column (headers_tracks[n].title, *cell);
          Gtk::TreeViewColumn *column = view_release->get_column (n);
          column->add_attribute (*cell, "text", n);
          column->set_resizable (true);
        }
        store_release = Gtk::ListStore::create (crtracks);
        view_release->set_model (store_release);
        view_release->get_selection()->set_mode (Gtk::SELECTION_NONE);

        // Local
        {
          Gtk::CellRendererToggle *cell = Gtk::manage ( new Gtk::CellRendererToggle() );
          view_local->append_column ("", *cell);
          Gtk::TreeViewColumn *column = view_local->get_column (0);
          column->add_attribute (*cell, "active", 0);
          column->set_resizable (false);
          cell->signal_toggled().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyTracks::on_mod_toggled));
        }

        for (unsigned int n = 0; n < G_N_ELEMENTS (headers_tracks); ++n)
        {
          Gtk::CellRendererText *cell = Gtk::manage ( new Gtk::CellRendererText() );
          cell->property_xalign() = headers_tracks[n].xalign;
          view_local->append_column (headers_tracks[n].title, *cell);
          Gtk::TreeViewColumn *column = view_local->get_column (n+1);
          if (n == 0)
            column->set_cell_data_func (*cell, sigc::mem_fun (*this, &Bmp::LibraryUIModifyTracks::cell_data_func_nr));
          else
            column->add_attribute (*cell, "text", n+1);
          column->set_resizable (true);
        }
        store_local = Gtk::ListStore::create (crupdatetracks);
        view_local->set_model (store_local);
        view_local->get_selection()->set_mode (Gtk::SELECTION_SINGLE);
        view_local->get_selection()->signal_changed().connect (sigc::mem_fun (*this, &Bmp::LibraryUIModifyTracks::local_selection_changed));
      }

      void
      LibraryUIModifyTracks::on_mod_toggled (const Glib::ustring& path)
      {
        Gtk::TreeModel::iterator m_Iter = store_local->get_iter (path);
        (*m_Iter)[crupdatetracks.mod] = !(*m_Iter)[crupdatetracks.mod];
      }

      void
      LibraryUIModifyTracks::cell_data_func_nr (Gtk::CellRenderer* cell_, const Gtk::TreeModel::iterator& iter)
      {
        Gtk::CellRendererText *cell = reinterpret_cast<Gtk::CellRendererText*>(cell_);
        static boost::format fint ("%d");
        unsigned int nr = (*iter)[crupdatetracks.nr];
        if (nr == 0)
          cell->property_text() = "";
        else
          cell->property_text() = (fint % nr).str();
      }

      void
      LibraryUIModifyTracks::local_selection_changed ()
      {
        bool selected = view_local->get_selection()->count_selected_rows ();
        b_move_up->set_sensitive (selected);
        b_move_down->set_sensitive (selected);
      }

      void
      LibraryUIModifyTracks::on_button_up ()
      {
        Gtk::TreeModel::iterator m_Iter = view_local->get_selection()->get_selected ();
        if (m_Iter == store_local->children().begin()) return;
        Gtk::TreeModel::iterator d_Iter = m_Iter;
        d_Iter--; // yes, this is slow, but it doesn't matter much as a release doesn't have that many tracks
        store_local->iter_swap (m_Iter, d_Iter);
      }

      void
      LibraryUIModifyTracks::on_button_down ()
      {
        Gtk::TreeModel::iterator m_Iter = view_local->get_selection()->get_selected ();
        if (m_Iter == store_local->children().end()) return;
        Gtk::TreeModel::iterator d_Iter = m_Iter;
        d_Iter++;
        store_local->iter_swap (m_Iter, d_Iter);
      }

      LibraryUIModifyTracks*
      LibraryUIModifyTracks::create ()
      {
        const std::string path = Glib::build_filename
              (DATA_DIR, "glade" G_DIR_SEPARATOR_S "library_ui_modify_tracks.glade");
        Glib::RefPtr<Gnome::Glade::Xml> glade_xml = Gnome::Glade::Xml::create (path);
        LibraryUIModifyTracks *library_ui = 0;
        glade_xml->get_widget_derived ("dialog_modify_tracks", library_ui);
        return library_ui;
      }

      int
      LibraryUIModifyTracks::run (const Glib::ustring& mb_albumid, 
				                          const Glib::ustring& artist,
                                  const Glib::ustring& album)
      {
          using namespace Bmp::DB;
          using namespace Bmp::Library;
          using namespace Bmp::MusicBrainzXML;
          using namespace Gtk;

          // Release
          MbTracks tracks; 
          get_tracks (mb_albumid, tracks);
          for (MbTracks::iterator iter = tracks.begin (), end = tracks.end (); iter != end; ++iter)
            {
              MbTrack const& track = *iter;
              TreeModel::iterator m_Iter  = store_release->append ();

              Glib::ustring title;

              if (track.artist)
                {
                  title.append (track.artist.get()).append(" - ").append(track.title.c_str());
                }
              else
                {
                  title = track.title.c_str(); 
                }

              (*m_Iter)[crtracks.nr]	      = track.tracknum; 
              (*m_Iter)[crtracks.title]	    = title; 
              (*m_Iter)[crtracks.track_id]  = track.track_id;

              if (track.artist_id)
                {
                  (*m_Iter)[crtracks.artist_id]  = track.artist_id.get();
                }
            }
          view_release->columns_autosize ();

          typedef Gtk::TreeNodeChildren Nodes;
          // Local
          std::string id_artist = library->get_metadatum_id (DATUM_ARTIST);
          std::string id_album  = library->get_metadatum_id (DATUM_ALBUM);

          AttributeList attrs; 
          VDataRow vector;

          attrs.push_back (Attribute (true, id_artist, artist));
          attrs.push_back (Attribute (true, id_album,  album));
          library->query (attrs, vector);

          for (VDataRow::const_iterator v = vector.begin (), end = vector.end (); v != end; ++v) 
            { 
              TreeModel::iterator m_Iter = store_local->append ();
              Track track (*v);
              if (track.title)	      (*m_Iter)[crupdatetracks.title] = track.title.get (); 
              if (track.tracknumber)  (*m_Iter)[crupdatetracks.nr] = track.tracknumber.get (); else (*m_Iter)[crupdatetracks.nr] = 0;
              if (track.location)	    (*m_Iter)[crupdatetracks.location] = track.location.get (); 
              (*m_Iter)[crupdatetracks.mod] = true;
            }
          view_local->columns_autosize ();

          Gtk::TreeNodeChildren const& nodes_release = store_release->children();
          Gtk::TreeNodeChildren const& nodes_local   = store_local->children();

          if (nodes_local.size () < nodes_release.size())
            {
              while (nodes_local.size () < nodes_release.size ())
              {
                store_local->append();
              }
            }

          try
            {
              Glib::RefPtr<Gdk::Pixbuf> cover;
              Amazon::get_cover_by_asin (library->get_asin_for_album (album), cover);
              i_coverart->set (cover);
            }
          catch (...)
            { 
              i_coverart->set (disc);
            }

          Glib::ustring m_artist;
          Glib::ustring m_album;

          m_artist.append("<b>").append(artist).append("</b>");
          m_album.append("<i>").append(album).append("</i>");

          l_artist->set_markup (m_artist);
          l_album->set_markup (m_album);

          int response = Gtk::Dialog::run ();

          if (response != GTK_RESPONSE_CANCEL) 
            {

              Nodes::const_iterator iter_r = nodes_release.begin ();
              Nodes::const_iterator iter_l = nodes_local.begin ();

              unsigned int max = nodes_local.size ();
              if (nodes_release.size() < nodes_local.size ()) max = nodes_release.size ();

              for (unsigned int n = 0; n < max; ++n)
                {
                  Glib::ustring location = (*iter_l)[crupdatetracks.location];
                  bool mod = (*iter_l)[crupdatetracks.mod];

                  if (!location.empty() && mod)
                    {
                      UpdateTrack track;
                      track.location	        = location; 
                      track.title	            = (*iter_r)[crtracks.title];
                      track.mb_artistid	      = (*iter_r)[crtracks.artist_id];
                      track.mb_trackid	      = (*iter_r)[crtracks.track_id];
                      track.tracknumber	      = (*iter_r)[crtracks.nr];
                      u_tracks.push_back (track); 
                    }
                  ++iter_r;
                  ++iter_l;
                }
            }
          return response;
      }
}
