//  BMPx - The Dumb Music Player
//  Copyright (C) 2005 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_MBWS_HPP 
#define BMP_MBWS_HPP

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glibmm.h>
#include <glibmm/markup.h>

#include <boost/format.hpp>
#include <boost/optional.hpp>

namespace Bmp
{
  namespace MusicBrainzXML
  {
      //////////////////// Release
      struct MbRelease
      { 
        Glib::ustring	release_id;
        Glib::ustring	release_name;

        Glib::ustring	asin;

        // FIXME: This is currently the 'release artist'
        //        In future, we should query a list of all tracks,
        //        so that an MbRelease also contains the MB
        //        metadata for each track

        Glib::ustring	artist_id;
        Glib::ustring	artist_name;

        unsigned int	score;
        unsigned int	disc_list;
        unsigned int	track_list;
        MbRelease () : score (0), disc_list (0), track_list (0) {}
      };
      typedef std::vector< MbRelease > MbReleases;

      class MbReleaseParser
        : public Glib::Markup::Parser
      {
          public:
            MbReleaseParser () : state (0) {}
            virtual ~MbReleaseParser () {}
            MbReleases releases;
          protected:
            void
            on_start_element  (Glib::Markup::ParseContext& context,
                   const Glib::ustring& element_name,
                   const AttributeMap& attributes);
            void
            on_end_element    (Glib::Markup::ParseContext& context,
                   const Glib::ustring& element_name);
            void
            on_text	      (Glib::Markup::ParseContext& context,
                   const Glib::ustring& text);
            void
            on_passtrough     (Glib::Markup::ParseContext& context,
                   const Glib::ustring& text);
            void
            on_error	      (Glib::Markup::ParseContext& context,
                   const Glib::MarkupError& error);
          private:
            enum Element
            {
              E_NONE	      = 1 << 0,
              E_RELEASE	    = 1 << 1,
              E_TITLE	      = 1 << 2,
              E_ASIN	      = 1 << 3,
              E_ARTIST	    = 1 << 4,
              E_NAME	      = 1 << 5,
              E_DISC_LIST   = 1 << 6,
              E_TRACK_LIST  = 1 << 7,
            };
            MbRelease    current;
            int           state;
            };
            void
            get_releases (const Glib::ustring& artist,
              const Glib::ustring& album,
              MbReleases& releases);
        
    }
}

namespace Bmp
{
  namespace MusicBrainzXML
  {
      //////////////////// Tracks for a Release
      struct MbTrack
      { 
        Glib::ustring	title;
        Glib::ustring	track_id;

        boost::optional<Glib::ustring>	artist;
        boost::optional<Glib::ustring>  artist_id;

        int tracknum;
        MbTrack () : tracknum (0) {};
      };
      typedef std::vector< MbTrack > MbTracks;

      class MbTrackParser
        : public Glib::Markup::Parser
      {
          public:
            MbTrackParser () : state (0), tracknum (1) {}
            virtual ~MbTrackParser () {}
            MbTracks tracks;
          protected:
            void
            on_start_element  (Glib::Markup::ParseContext& context,
                   const Glib::ustring& element_name,
                   const AttributeMap& attributes);
            void
            on_end_element    (Glib::Markup::ParseContext& context,
                   const Glib::ustring& element_name);
            void
            on_text	      (Glib::Markup::ParseContext& context,
                   const Glib::ustring& text);
            void
            on_passtrough     (Glib::Markup::ParseContext& context,
                   const Glib::ustring& text);
            void
            on_error	      (Glib::Markup::ParseContext& context,
                   const Glib::MarkupError& error);
          private:

              enum Element
              {
                E_NONE	    = 1 << 0,
                E_TRACK	    = 1 << 1,
                E_TITLE	    = 1 << 2,
                E_ARTIST	  = 1 << 3,
                E_NAME	    = 1 << 4,
              };

              MbTrack       current;
              int           state;
              int		  tracknum;

            };

            void
            get_tracks   (const Glib::ustring& mb_albumid,
                          MbTracks& tracks);
  }
}

#endif
